package com.qmth.teachcloud.mark.service.impl;

import com.alibaba.fastjson.JSON;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.metadata.OrderItem;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.qmth.boot.api.exception.ApiException;
import com.qmth.teachcloud.common.contant.SystemConstant;
import com.qmth.teachcloud.common.entity.BasicCourse;
import com.qmth.teachcloud.common.entity.MarkQuestion;
import com.qmth.teachcloud.common.entity.SysUser;
import com.qmth.teachcloud.common.enums.ExceptionResultEnum;
import com.qmth.teachcloud.common.enums.mark.*;
import com.qmth.teachcloud.common.service.BasicCourseService;
import com.qmth.teachcloud.common.service.BasicOperationLogService;
import com.qmth.teachcloud.common.service.SysUserService;
import com.qmth.teachcloud.common.util.ServletUtil;
import com.qmth.teachcloud.mark.dto.mark.ScoreItem;
import com.qmth.teachcloud.mark.dto.mark.manage.Task;
import com.qmth.teachcloud.mark.dto.mark.mark.MarkSettingDto;
import com.qmth.teachcloud.mark.dto.mark.mark.MarkStatusDto;
import com.qmth.teachcloud.mark.dto.mark.mark.SubmitResult;
import com.qmth.teachcloud.mark.entity.*;
import com.qmth.teachcloud.mark.enums.ExamType;
import com.qmth.teachcloud.mark.enums.LockType;
import com.qmth.teachcloud.mark.enums.MarkTaskStatus;
import com.qmth.teachcloud.mark.enums.QuestionModel;
import com.qmth.teachcloud.mark.lock.LockService;
import com.qmth.teachcloud.mark.params.MarkHeaderGroupResult;
import com.qmth.teachcloud.mark.params.MarkResult;
import com.qmth.teachcloud.mark.params.MarkResultQuestion;
import com.qmth.teachcloud.mark.service.*;
import com.qmth.teachcloud.mark.utils.TaskLock;
import com.qmth.teachcloud.mark.utils.TaskLockUtil;
import io.lettuce.core.GeoArgs.Sort;
import org.apache.commons.collections4.CollectionUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

@Service
public class MarkServiceImpl implements MarkService {

    private final static Logger log = LoggerFactory.getLogger(MarkServiceImpl.class);

    private Map<Long, Long> markerLastUpdateTime = new ConcurrentHashMap<>();

    @Resource
    private BasicCourseService basicCourseService;
    @Resource
    private MarkPaperService markPaperService;

    @Resource
    private MarkQuestionService markQuestionService;
    @Resource
    private MarkUserPaperService markUserPaperService;
    @Resource
    private MarkUserQuestionService markUserQuestionService;
    @Resource
    private MarkUserClassService markUserClassService;

    @Resource
    private MarkTaskService markTaskService;

    @Resource
    private MarkStudentService markStudentService;

    @Resource
    private MarkProblemHistoryService markProblemHistoryService;

    @Resource
    private MarkArbitrateHistoryService markArbitrateHistoryService;

    @Resource
    private MarkSubjectiveScoreService markSubjectiveScoreService;
    @Resource
    LockService lockService;

    @Resource
    ScanAnswerCardService scanAnswerCardService;

    @Resource
    TaskService taskService;

    @Resource
    private MarkHeaderHistoryService markHeaderHistoryService;
    @Resource
    private SysUserService sysUserService;
    @Resource
    private MarkRejectHistoryService markRejectHistoryService;
    @Resource
    private BasicOperationLogService basicOperationLogService;

    /**
     * 释放某个评卷员已完成的评卷任务
     *
     * @param submitResult 评卷结果
     * @param userId       评卷员
     */
    @Override
    public void releaseTask(SubmitResult submitResult, Long userId) {
        if (submitResult.getStudentId() != null) {
            TaskLock taskLock = TaskLockUtil.getFormalTask(String.valueOf(submitResult.getStudentId()));
            taskLock.remove(submitResult.getStudentId(), 1, userId);
            taskLock.refresh(userId);
        }
    }

    /**
     * 某个评卷分组已申请的评卷任务数量
     *
     * @return
     */
    @Override
    public int applyCurrentCount(MarkQuestion markQuestion) {
        // todo 正在评卷数量查询 (评卷员和评卷分组分开2个方法)
        TaskLock taskLock = getTaskLock(markQuestion);
        int count = 0;
        if (taskLock != null) {
            count = taskLock.count();
        }
        return count;
    }

    private TaskLock getTaskLock(MarkQuestion markQuestion) {
        MarkPaper markPaper = markPaperService.getByExamIdAndPaperNumber(markQuestion.getExamId(),
                markQuestion.getPaperNumber());
        if (markPaper.getStatus() == MarkPaperStatus.FORMAL) {
            return TaskLockUtil.getFormalTask(getGroupKey(markQuestion));
        }
        return null;
    }

    /**
     * 释放某个评卷员的所有锁定任务
     *
     * @param markUserGroup
     */
    @Override
    public void releaseByMarkUserGroup(MarkUserQuestion markUserGroup) {
        TaskLock taskLock = TaskLockUtil.getFormalTask(getGroupKey(markUserGroup));
        taskLock.clear(markUserGroup.getUserId());
    }

    @Override
    public int applyCurrentCount(MarkQuestion markQuestion, Long markUserGroupId) {
        int count = 0;
        TaskLock taskLock = getTaskLock(markQuestion);
        if (taskLock != null) {
            count = taskLock.count(markUserGroupId);
        }
        return count;
    }

    @Override
    public void resetMarker(MarkUserQuestion markUserQuestion) {
        Long examId = markUserQuestion.getExamId();
        String paperNumber = markUserQuestion.getPaperNumber();
        Long questionId = markUserQuestion.getQuestionId();
        Long userId = markUserQuestion.getUserId();
        MarkPaper markPaper = markPaperService.getByExamIdAndPaperNumber(examId, paperNumber);
        if (markPaper == null || markUserQuestionService.getById(markUserQuestion.getId()) == null) {
            return;
        }
        if (markPaper.getStatus() == MarkPaperStatus.FORMAL) {
            // 遍历相关评卷任务的模式
            List<MarkTaskStatus> statusList = Arrays.asList(MarkTaskStatus.ARBITRATED, MarkTaskStatus.WAIT_ARBITRATE,
                    MarkTaskStatus.PROBLEM);
            List<MarkTask> markTaskList = markTaskService.listByExamIdAndPaperNumberAndQuestionIdAndUserIdAndStatusNotIn(examId, paperNumber, questionId, userId, statusList);
            for (MarkTask markTask : markTaskList) {
                Long studentId = markTask.getStudentId();
                markTaskService.resetById(markTask.getId(), null, null, null, null, MarkTaskStatus.WAITING);
                lockService.waitlock(LockType.STUDENT, markTask.getStudentId());
                markStudentService.updateSubjectiveStatusAndScore(studentId, SubjectiveStatus.UNMARK, null, null);
                lockService.unlock(LockType.STUDENT, markTask.getStudentId());
            }
            markUserQuestionService.resetById(markUserQuestion.getId());
        }
        this.updateMarkedCount(examId, paperNumber, questionId);
        releaseByMarkUserGroup(markUserQuestion);
    }

    @Override
    public void updateMarkedCount(Long examId, String paperNumber, Long questionId) {
        List<MarkTaskStatus> markTaskStatuses = Arrays.asList(MarkTaskStatus.MARKED, MarkTaskStatus.ARBITRATED);
        //查询已评和已仲裁的任务数
        int count = markTaskService.countByExamIdAndPaperNumberAndQuestionIdAndStatusIn(examId, paperNumber,
                questionId, markTaskStatuses);
        //更新当前小题的已评数量
        markQuestionService.updateMarkedCount(questionId, count);
    }

    @Override
    public boolean rejectMarkTask(MarkProblemHistory markProblemHistory, MarkTask markTask, Long userId) {
        MarkPaper markPaper = markPaperService.getByExamIdAndPaperNumber(markTask.getExamId(), markTask.getPaperNumber());
        if (markPaper.getStatus() == MarkPaperStatus.FINISH) {
            return false;
        }
        Long now = System.currentTimeMillis();
        if (markTaskService.resetById(markTask.getId(), null, null, userId, now, MarkTaskStatus.WAITING)) {
            resetStudentGroup(markTask.getStudentId());
            markProblemHistoryService.resetByMarkProblemId(markProblemHistory.getId(), MarkProblemStatus.WAITING,
                    userId, MarkProblemStatus.BACK, now);
            updateMarkedCount(markTask.getExamId(), markTask.getPaperNumber(), markTask.getQuestionId());
            return true;
        } else {
            return false;
        }
    }

    private void resetStudentGroup(Long studentId) {
        markStudentService.updateSubjectiveStatusAndScore(studentId, SubjectiveStatus.UNMARK, null, null);
    }


    @Override
    public String getGroupKey(MarkQuestion markQuestion) {
//        return markQuestion.getExamId() + "_" + markQuestion.getPaperNumber() + "_" + markQuestion.getId();
        return String.valueOf(markQuestion.getId());
    }

    @Override
    public void updateQuality(MarkUserQuestion markUserQuestion) {
        List<MarkTask> list = markTaskService.listByExamIdAndPaperNumberAndQuestionIdAndUserIdAndClassName(
                markUserQuestion.getExamId(), markUserQuestion.getPaperNumber(), markUserQuestion.getQuestionId(),
                markUserQuestion.getUserId(), null);
        int finishCount = 0;
        int validCount = 0;
        int headerFinishCount = 0;
        double sumScore = 0;
        double sumScore2 = 0;
        double avgScore = 0;
        double stdevScore = 0;
        double sumSpent = 0;
        double avgSpent = 0;
        double maxScore = 0;
        double minScore = 0;
        for (MarkTask markTask : list) {
            if (!markTask.getStatus().equals(MarkTaskStatus.PROBLEM)) {
                finishCount++;
            }
            if (markTask.getStatus() == MarkTaskStatus.MARKED && markTask.getHeaderScore() == null) {
                validCount++;
            }
            // 管理员打分数量（主观题检查，仲裁）
            if ((markTask.getStatus() == MarkTaskStatus.MARKED || markTask.getStatus() == MarkTaskStatus.ARBITRATED) && markTask.getHeaderScore() != null) {
                headerFinishCount++;
            }
            double score = markTask.getMarkerScore() != null ? markTask.getMarkerScore() : 0;
            int spent = markTask.getMarkerSpent() != null ? markTask.getMarkerSpent() : 0;

            sumScore += score;
            sumScore2 += Math.pow(score, 2);
            sumSpent += spent;
            if (finishCount > 0) {
                avgScore = sumScore / finishCount;
                avgSpent = sumSpent / finishCount;
                // 递归法计算标准差
                stdevScore = Math.sqrt(sumScore2 / finishCount - Math.pow(sumScore / finishCount, 2));
            }

            // 最高分
            if (score > maxScore) {
                maxScore = score;
            }
            // 最低分
            if (finishCount == 1) {
                minScore = score;
            } else if (score < minScore) {
                minScore = score;
            }
        }
        markUserQuestionService.updateQualityById(markUserQuestion.getId(), finishCount, headerFinishCount, validCount,
                avgSpent / 1000, avgScore, stdevScore, maxScore, minScore);
        markerLastUpdateTime.put(markUserQuestion.getId(), System.currentTimeMillis());
    }

    @Override
    public boolean needUpdateQuality(MarkUserQuestion markUserQuestion, int expireMinutes) {
        if (markUserQuestion == null) {
            return false;
        }
        Long lastUpdateTime = markerLastUpdateTime.get(markUserQuestion.getId());
        MarkTask markTask = markTaskService.getLastOneByUserIdAndStatus(markUserQuestion.getExamId(),
                markUserQuestion.getPaperNumber(), markUserQuestion.getQuestionId(), markUserQuestion.getUserId(),
                MarkTaskStatus.MARKED);
        if (markTask != null && markTask.getMarkerTime() != null) {
            long lastMarkTime = markTask.getMarkerTime();
            if (lastUpdateTime != null && lastUpdateTime > lastMarkTime) {
                return false;
            } else {
                return (System.currentTimeMillis() - lastMarkTime) < (expireMinutes * 60 * 1000);
            }
        }
        return false;
    }

    /**
     * 管理员/组长处理仲裁卷
     *
     * @param markResult
     * @param userId
     */
    @Transactional
    @Override
    public void processArbitrate(MarkHeaderGroupResult markResult, Long userId) {
        MarkArbitrateHistory markArbitrateHistory = markArbitrateHistoryService.getById(markResult.getTaskId());
        markArbitrateHistory.setUpdateUserId(userId);
        markArbitrateHistory.setTotalScore(markResult.getScore());
//        markArbitrateHistory.setScoreList(markResult.isUnselective() ? null : markResult.getScoreList());
        markArbitrateHistory.setStatus(MarkArbitrateStatus.MARKED);
        markArbitrateHistory.setUpdateTime(System.currentTimeMillis());
//        // 保存阅卷轨迹
//        if (markResult.getTrackList() != null && !markResult.isUnselective()) {
//            markHeaderTrackService.deleteByExamIdAndPaperNumberAndGroupNumberAndStudentId(
//                    markArbitrateHistory.getExamId(), markArbitrateHistory.getPaperNumber(),
//                    markArbitrateHistory.getGroupNumber(), markArbitrateHistory.getStudentId());
//            List<MarkHeaderTrack> tracks = markResult.getTrackList(markArbitrateHistory);
//            markHeaderTrackService.saveOrUpdateBatchByMultiId(tracks);
//        }
//        // 保存特殊标记
//        if (markResult.getSpecialTagList() != null) {
//            markHeaderTagService.deleteByStudentIdAndGroupNumber(markArbitrateHistory.getStudentId(),
//                    markArbitrateHistory.getGroupNumber());
//            markHeaderTagService.saveBatch(markResult.getHeaderTagList(markArbitrateHistory));
//        }
        markArbitrateHistoryService.saveOrUpdate(markArbitrateHistory);
        markTaskService.updateHeaderResult(markArbitrateHistory.getExamId(), markArbitrateHistory.getPaperNumber(),
                markArbitrateHistory.getQuestionId(), markArbitrateHistory.getStudentId(),
                markArbitrateHistory.getUpdateUserId(), markResult.getScore(), markResult.getTrackList(), markResult.getSpecialTagList(), markArbitrateHistory.getUpdateTime(), MarkTaskStatus.ARBITRATED);
        updateMarkedCount(markArbitrateHistory.getExamId(), markArbitrateHistory.getPaperNumber(),
                markArbitrateHistory.getQuestionId());
        checkStudentQuestion(markArbitrateHistory.getStudentId(), markQuestionService.getById(markArbitrateHistory.getQuestionId()), null);
        // 评卷质量重新统计
        List<MarkUserQuestion> markUserGroups = markUserQuestionService.listByExamIdAndPaperNumberAndQuestionId(markArbitrateHistory.getExamId(), markArbitrateHistory.getPaperNumber(), markArbitrateHistory.getQuestionId());
        markUserGroups.forEach(m -> this.updateQuality(m));
    }

    private String getGroupKey(MarkUserQuestion markUserGroup) {
        return markUserGroup.getExamId() + "_" + markUserGroup.getPaperNumber();
    }

    /**
     * 考生分组判断是否评卷完成，以及后续的统一处理动作
     *
     * @param studentId
     * @param markQuestion
     * @param result
     */
    private void checkStudentQuestion(Long studentId, MarkQuestion markQuestion, MarkResultQuestion result) {
        if (calculateQuestionId(markQuestion, studentId)) {
            //更新考生分组分数
            updateStudentQuestionScore(studentId, markQuestion, result);
//            checkStudentSubjective(studentId, markQuestion.getExamId(), markQuestion.getPaperNumber(), version);
        } else {
            markStudentService.updateSubjectiveStatusAndScore(studentId, SubjectiveStatus.UNMARK, null, null);
        }
    }

    @Override
    @Transactional
    public void checkStudentSubjective(Long studentId, Long examId, String paperNumber, Integer version) {
        long unBindMarkerQuestionCount = markQuestionService.countUnBindMarkerByExamIdAndPaperNumberAndObjective(examId, paperNumber, false);
        long questionCount = markQuestionService.countByExamIdAndPaperNumberAndObjective(examId, paperNumber, false);
        long subjectiveScoreCount = markSubjectiveScoreService.countByStudentId(studentId);
        // 主观题数大于0，主观题全部绑定了评卷员，考生小题分数数量等于主观题数量
        if (questionCount > 0 && unBindMarkerQuestionCount == 0 && questionCount == subjectiveScoreCount) {
            scoreCalculate(studentId, version);
        } else {//否则更新该学生主观题状态为未阅卷
//            markStudentService.updateSubjectiveStatusAndScore(studentId, SubjectiveStatus.UNMARK, null, null);
            markStudentService.updateSubjectiveScoreByVersion(studentId, SubjectiveStatus.MARKED, null, null, version);
        }
    }

    @Override
    public void buildMarkTask(MarkPaper markPaper) {
        try {
            lockService.watch(LockType.EXAM_SUBJECT, markPaper.getExamId(), markPaper.getPaperNumber());
            log.info("start create library for examId=" + markPaper.getExamId() + ", paperNumber="
                    + markPaper.getPaperNumber());
            // 清除缺考考生和违纪考生（违纪考生参与阅卷，update by 2024-10-17）
            List<MarkStudent> markStudentList = markStudentService
                    .listAbsentOrBreachMarkTaskStudent(markPaper.getExamId(), markPaper.getPaperNumber());
            if (CollectionUtils.isNotEmpty(markStudentList)) {
                for (MarkStudent student : markStudentList) {
                    try {
                        lockService.waitlock(LockType.STUDENT, student.getId());
                        log.info("delete library for studentId=" + student.getId());
                        this.deleteMarkTaskByStudent(student);
                    } catch (Exception e) {
                        log.error("delete student library error", e);
                    } finally {
                        lockService.unlock(LockType.STUDENT, student.getId());
                    }
                }
            }
            // 处理正常考生
            List<MarkQuestion> markQuestionList = markQuestionService.listByExamIdAndPaperNumberAndObjective(markPaper.getExamId(), markPaper.getPaperNumber(), false);
            for (MarkQuestion markQuestion : markQuestionList) {
                // 生成正评任务
                buildFormalLibrary(markQuestion);
            }
        } finally {
            lockService.unwatch(LockType.EXAM_SUBJECT, markPaper.getExamId(), markPaper.getPaperNumber());
        }
    }

    private void buildFormalLibrary(MarkQuestion markQuestion) {
        int pageSize = 100;
        try {
            lockService.watch(LockType.QUESTION, markQuestion.getId());
            // 上锁后重复验证分组状态
            MarkPaper markPaper = markPaperService.getByExamIdAndPaperNumber(markQuestion.getExamId(), markQuestion.getPaperNumber());
            if (MarkPaperStatus.FINISH.equals(markPaper.getStatus())) {
                return;
            }
            int count = 0;
            List<MarkStudent> studentList = markStudentService.listUnMarkTaskStudent(markQuestion.getExamId(), markQuestion.getPaperNumber(), markQuestion.getId(), pageSize);
            while (CollectionUtils.isNotEmpty(studentList)) {
                // 已生成的双评任务总数的第一组任务数
                int doubleMarkTaskCount1 = markTaskService.countByExamIdAndPaperNumberAndQuestionIdAndTaskNumber(
                        markQuestion.getExamId(), markQuestion.getPaperNumber(), markQuestion.getId(), 1);
                // 已生成的双评任务总数的第二组任务数
                int doubleMarkTaskCount2 = markTaskService.countByExamIdAndPaperNumberAndQuestionIdAndTaskNumber(
                        markQuestion.getExamId(), markQuestion.getPaperNumber(), markQuestion.getId(), 2);
                List<MarkTask> taskList = new ArrayList<>();
                int doubleCount = 0;
                int studentCount = studentList.size();
                for (MarkStudent student : studentList) {
                    MarkTask markTask = new MarkTask();
                    markTask.setId(SystemConstant.getDbUuid());
                    markTask.setExamId(student.getExamId());
                    markTask.setCourseId(student.getCourseId());
                    markTask.setPaperNumber(student.getPaperNumber());
                    markTask.setQuestionId(markQuestion.getId());
                    markTask.setMainNumber(markQuestion.getMainNumber());
                    markTask.setSubNumber(markQuestion.getSubNumber());
                    markTask.setStudentId(student.getId());
                    markTask.setBasicStudentId(student.getBasicStudentId());
                    markTask.setStudentCode(student.getStudentCode());
                    markTask.setSecretNumber(student.getSecretNumber());
                    markTask.setTaskNumber(1);
                    markTask.setStatus(MarkTaskStatus.WAITING);
                    taskList.add(markTask);
                    // 开启双评时需要判断是否生成第二份评卷任务
                    if (markQuestion.getDoubleRate() != null && markQuestion.getDoubleRate() > 0) {
                        boolean needDouble;
                        if (markQuestion.getDoubleRate() == 100) {
                            needDouble = true;
                        } else {
                            double libraryCount = taskList.size();
                            int expectCount = (int) ((doubleMarkTaskCount1 + studentCount) * markQuestion.getDoubleRate()
                                    / 100);
                            // 随机数判断加入当前已经生成双评任务的比例加权
                            // 实际双评任务数小于理论生成数 &&（剩余未生成双评的考生数量小于剩余应生成的数量）
                            needDouble = (doubleMarkTaskCount2 + doubleCount) < expectCount
                                    && ((studentCount - libraryCount + doubleCount) <= (expectCount
                                    - doubleMarkTaskCount2 - doubleCount));
                        }
                        if (needDouble) {
                            markTask = new MarkTask();
                            markTask.setId(SystemConstant.getDbUuid());
                            markTask.setExamId(student.getExamId());
                            markTask.setCourseId(student.getCourseId());
                            markTask.setPaperNumber(student.getPaperNumber());
                            markTask.setQuestionId(markQuestion.getId());
                            markTask.setMainNumber(markQuestion.getMainNumber());
                            markTask.setSubNumber(markQuestion.getSubNumber());
                            markTask.setStudentId(student.getId());
                            markTask.setBasicStudentId(student.getBasicStudentId());
                            markTask.setStudentCode(student.getStudentCode());
                            markTask.setSecretNumber(student.getSecretNumber());
                            markTask.setTaskNumber(2);
                            markTask.setStatus(MarkTaskStatus.WAITING);
                            taskList.add(markTask);
                            doubleCount++;
                        }
                    }
                }
                // 有新任务创建
                if (CollectionUtils.isNotEmpty(taskList)) {
                    count += taskList.size();
                    // 任务乱序
                    Collections.shuffle(taskList);
                    // 批量保存
                    markTaskService.saveBatch(taskList);
                    // 正评状态才需要更新任务数量
                    if (MarkPaperStatus.FORMAL.equals(markPaper.getStatus())) {
                        this.updateMarkTaskCount(markQuestion.getExamId(), markQuestion.getPaperNumber(), markQuestion.getId());
                        this.updateMarkedCount(markQuestion.getExamId(), markQuestion.getPaperNumber(), markQuestion.getId());
                    }
                }
                studentList = markStudentService.listUnMarkTaskStudent(markQuestion.getExamId(), markQuestion.getPaperNumber(), markQuestion.getId(), pageSize);
            }
            if (count != 0) {
                log.info("finish create " + count + " markTask for examId=" + markQuestion.getExamId() + ", paperNumber="
                        + markQuestion.getPaperNumber() + ", questionId=" + markQuestion.getId());
            }
        } catch (Exception e) {
            log.error("build formal markTask error for examId=" + markQuestion.getExamId() + ", paperNumber="
                    + markQuestion.getPaperNumber() + ", questionId=" + markQuestion.getId(), e);
        } finally {
            lockService.unwatch(LockType.QUESTION, markQuestion.getId());
        }
    }

    @Transactional
    @Override
    public void deleteMarkTaskByStudent(MarkStudent student) {
        markStudentService.updateSubjectiveStatusAndScore(student.getId(), SubjectiveStatus.UNMARK, null, null);
        // 正评相关数据
        markArbitrateHistoryService.deleteByStudentId(student.getId());
        markProblemHistoryService.deleteByStudentId(student.getId());
        markTaskService.deleteByStudentId(student.getId());
        // 主观状态与得分明细
        markSubjectiveScoreService.deleteByStudentId(student.getId());
        updateGroupAllCount(student.getExamId(), student.getPaperNumber());
        // 复核记录
        // inspectedService.clearByStudent(student.getId());
        // 打回记录
        // rejectHistoryDao.deleteByStudentId(student.getId());
    }

    @Transactional
    @Override
    public void updateGroupAllCount(Long examId, String paperNumber) {
        List<MarkQuestion> markQuestionList = markQuestionService.listByExamIdAndPaperNumberAndObjective(examId, paperNumber, false);
        for (MarkQuestion markQuestion : markQuestionList) {
            updateMarkedCount(markQuestion.getExamId(), markQuestion.getPaperNumber(), markQuestion.getId());
            updateMarkTaskCount(markQuestion.getExamId(), markQuestion.getPaperNumber(), markQuestion.getId());
        }
    }

    private void updateMarkTaskCount(Long examId, String paperNumber, Long questionId) {
        int count = markTaskService.countByExamIdAndPaperNumberAndQuestionIdAndUserId(examId, paperNumber, questionId,
                null);
        markQuestionService.updateTaskCount(questionId, count);
    }

    private boolean calculateQuestionId(MarkQuestion markQuestion, Long studentId) {
        List<MarkTask> list = markTaskService.listByExamIdAndPaperNumberAndQuestionIdAndStudentId(markQuestion.getExamId(),
                markQuestion.getPaperNumber(), markQuestion.getId(), studentId);
        if (list.isEmpty()) {
            return false;
        }
        for (MarkTask markTask : list) {
            if (markTask.getStatus() != MarkTaskStatus.MARKED && markTask.getStatus() != MarkTaskStatus.ARBITRATED) {
                // 有非完成状态的评卷任务，直接返回
                return false;
            }
        }
        return true;
    }

    private void updateStudentQuestionScore(Long studentId, MarkQuestion markQuestion, MarkResultQuestion result) {
        MarkSubjectiveScore ss = markSubjectiveScoreService.getByStudentIdAndQuestionId(studentId, markQuestion.getId());
        if (ss == null) {
            ss = new MarkSubjectiveScore();
        }
        ss.setStudentId(studentId);
        ss.setExamId(markQuestion.getExamId());
        ss.setPaperNumber(markQuestion.getPaperNumber());
        ss.setQuestionId(markQuestion.getId());
        ss.setMainNumber(markQuestion.getMainNumber());
        ss.setSubNumber(markQuestion.getSubNumber());
        ss.setScore(result.getMarkerScore());
        ss.setMainScore(0.0);
        markSubjectiveScoreService.saveOrUpdateByMultiId(ss);
    }

    private void scoreCalculate(Long studentId, Integer version) {
        List<ScoreItem> scoreList = new ArrayList<>();
        Map<Integer, List<MarkSubjectiveScore>> mainScoreMap = new HashMap<>();
        Map<Integer, Double> scoreMap = new HashMap<>();

        // 循环所有主观得分明细
        List<MarkSubjectiveScore> list = markSubjectiveScoreService.listByStudentId(studentId);
        // list.sort(null);
        for (MarkSubjectiveScore ss : list) {
            List<MarkSubjectiveScore> mainScoreList = mainScoreMap.get(ss.getMainNumber());
            if (mainScoreList == null) {
                mainScoreList = new ArrayList<>();
            }
            mainScoreList.add(ss);
            mainScoreMap.put(ss.getMainNumber(), mainScoreList);
            scoreList.add(new ScoreItem(ss));
        }
        // 计算大题分
        for (Integer mainNumber : mainScoreMap.keySet()) {
            List<MarkSubjectiveScore> mainScoreList = mainScoreMap.get(mainNumber);
            BigDecimal mainScore = BigDecimal.ZERO;
            for (MarkSubjectiveScore subjectiveScore : mainScoreList) {
                mainScore = mainScore.add(BigDecimal.valueOf(subjectiveScore.getScore()));
            }
            for (MarkSubjectiveScore subjectiveScore : mainScoreList) {
                subjectiveScore.setMainScore(mainScore.doubleValue());
                markSubjectiveScoreService.saveOrUpdateByMultiId(subjectiveScore);
            }
            scoreMap.put(mainNumber, mainScore.doubleValue());
        }
        // 计算选做题分数
        BigDecimal totalScore = BigDecimal.ZERO;
        // 计算非选做题总分
        for (Integer mainNumber : scoreMap.keySet()) {
            totalScore = totalScore.add(BigDecimal.valueOf(scoreMap.get(mainNumber)));
        }
        // 全部评完，更新考生主观题得分
//        markStudentService.updateSubjectiveStatusAndScore(studentId, SubjectiveStatus.MARKED, totalScore.doubleValue(),
//                MarkStudent.buildScoreList(scoreList));
        markStudentService.updateSubjectiveScoreByVersion(studentId, SubjectiveStatus.MARKED, totalScore.doubleValue(), MarkStudent.buildScoreList(scoreList), version);
    }

    @Override
    public MarkSettingDto getSetting(SysUser user, Long examId, String paperNumber) {
        MarkUserPaper markUserPaper = markUserPaperService.getByExamIdAndPaperNumberAndUserId(examId, paperNumber, user.getId());
        user = sysUserService.getById(user.getId());
        MarkPaper markPaper = markPaperService.getByExamIdAndPaperNumber(examId, paperNumber);

        MarkSettingDto dto = new MarkSettingDto();
        dto.setExamType(ExamType.SCAN_IMAGE);//考试类型，默认SCAN_IMAGE
        dto.setMode(markUserPaper != null && markUserPaper.getMode() != null ? markUserPaper.getMode()
                : MarkMode.UNLIMITED.equals(markPaper.getMarkMode()) ? MarkMode.TRACK : markPaper.getMarkMode());//模式 TRACK/COMMON
        dto.setQuestionModel(markUserPaper != null && markUserPaper.getQuestionModel() != null ? markUserPaper.getQuestionModel() : QuestionModel.MULTI);
        dto.setForceMode(markPaper.getForceMode());//强制模式切换
        dto.setSheetView(markPaper.getSheetView());//是否显示原图功能
        dto.setSheetConfig(markPaper.getSheetConfig());//原图遮盖规则
        dto.setEnableAllZore(false);//是否开启全零分（知学知考阅卷默认false）
        dto.setFileServer(null);//图片服务地址
        dto.setUserName(user.getRealName());//评卷员名称
        dto.getSubject().setAnswerUrl(markQuestionService.previewAnswerFileByExamIdAndPaperNumber(examId, paperNumber));
        dto.getSubject().setPaperUrl(markQuestionService.previewPaperFileByExamIdAndPaperNumber(examId, paperNumber));
        BasicCourse basicCourse = basicCourseService.getById(markPaper.getCourseId());
        dto.getSubject().setCode(markPaper.getPaperNumber());
        dto.getSubject().setName(basicCourse != null ? basicCourse.getName() : null);
        dto.setForceSpecialTag(false);//强制标记是否开启 必须要有标记(分数轨迹或特殊标记)（知学知考阅卷默认true）
        dto.setUiSetting(user.getUiSetting());//新交互模式下，以下字段要重新定义或新增(初始值为空，按评卷员保存)
        dto.setStatusValue(markPaper.getStatus());//只显示试评名称 FORMAL("正评"), FINISH("结束")
        dto.setProblemTypes(MarkProblemType.listTypes());//问题卷类型
//        dto.setTopCount(markUserGroup != null && markUserGroup.getTopCount() != null ? markUserGroup.getTopCount() : 0);
        dto.setSplitConfig(new Double[0]);//使用裁切整图时的裁切配置
        dto.setPrefetchCount(3);//预加载任务数量
        dto.setStartTime(markPaper.getMarkStartTime());//评卷开始时间
        dto.setEndTime(markPaper.getMarkEndTime());//评卷结束时间
        dto.setSelective(false);//是否为选做题（默认false）
        dto.setAutoScroll(markPaper.getAutoScroll());//是否自动跳转
        dto.setEnableSplit(false);//是否裁切（默认false）
        dto.setShowObjectScore(markPaper.getShowObjectScore());//是否显示客观分
        return dto;
    }

    @Override
    public MarkStatusDto getStatus(Long userId, Long examId, String paperNumber) {
        MarkStatusDto dto = new MarkStatusDto();
        MarkPaper markPaper = markPaperService.getByExamIdAndPaperNumber(examId, paperNumber);

        // 评卷方式
        MarkUserPaper markUserPaper = markUserPaperService.getByExamIdAndPaperNumberAndUserId(examId, paperNumber, userId);
        Long questionId = null;
        if (markUserPaper != null && markUserPaper.getQuestionModel().equals(QuestionModel.SINGLE)) {
            questionId = markTaskService.minQuestionIdByExamIdAndPaperNumber(examId, paperNumber, MarkTaskStatus.MARKED, MarkTaskStatus.ARBITRATED);
        }
        List<String> classNames = null;
        //校验是否有分班阅
        if (markPaper != null && markPaper.getClassMark() != null && markPaper.getClassMark().booleanValue()) {
            List<MarkUserClass> markUserClassList = markUserClassService.listByExamIdAndPaperNumberAndUserId(examId, paperNumber, userId);
            if (CollectionUtils.isNotEmpty(markUserClassList)) {
                classNames = markUserClassList.stream().map(MarkUserClass::getClassName).collect(Collectors.toList());
            }
        }
        //待仲裁卷数量
        dto.setArbitrateCount(markTaskService.countByExamIdAndPaperNumberAndQuestionIdAndUserIdAndAndClassNameStatusIn(examId,
                paperNumber, null, null, classNames, MarkTaskStatus.WAIT_ARBITRATE));
        //总数量
        dto.setTotalCount(markTaskService.countByExamIdAndPaperNumberAndQuestionIdAndUserIdAndAndClassNameStatusIn(examId, paperNumber, questionId, null, classNames));
        //总评卷数量（已评）
        dto.setMarkedCount(markTaskService.countByExamIdAndPaperNumberAndQuestionIdAndUserIdAndAndClassNameStatusIn(examId, paperNumber, questionId, null, classNames, MarkTaskStatus.MARKED, MarkTaskStatus.ARBITRATED));
        //个人评卷数量
        dto.setPersonCount(markTaskService.countByExamIdAndPaperNumberAndQuestionIdAndUserIdAndAndClassNameStatusIn(examId,
                paperNumber, questionId, userId, classNames, MarkTaskStatus.MARKED, MarkTaskStatus.ARBITRATED,
                MarkTaskStatus.WAIT_ARBITRATE));
        //问题卷数量
        dto.setProblemCount(markTaskService.countByExamIdAndPaperNumberAndQuestionIdAndUserIdAndAndClassNameStatusIn(examId,
                paperNumber, questionId, null, classNames, MarkTaskStatus.PROBLEM));
        return dto;
    }

    @Override
    public void clear(Long userId, Long examId, String paperNumber) {
        String key = examId + "_" + paperNumber;
        TaskLock taskLock = TaskLockUtil.getFormalTask(key);
        taskLock.clear(userId);
    }

    @Override
    public void releaseByStudent(MarkStudent student) {
        TaskLock taskLock = TaskLockUtil.getInspectedStudentTask(getKey(student));
        taskLock.remove(student.getId(), 1);
    }

    @Override
    public boolean applyStudent(MarkStudent student, Long userId) {
        TaskLock taskLock = TaskLockUtil.getInspectedStudentTask(getKey(student));
        boolean lock = taskLock.add(student.getId(), 1, userId);
        // 上锁失败直接返回
        if (!lock) {
            return false;
        }
        // 重复校验任务状态
        if (student.getSubjectiveStatus().equals(SubjectiveStatus.MARKED)) {
            return true;
        } else {
            taskLock.remove(student.getId(), 1, userId);
            return false;
        }
    }

    @Override
    public void submitHeaderTask(List<MarkResultQuestion> resultQuestions, MarkStudent markStudent) {
        Long userId = ServletUtil.getRequestUserId();
        for (MarkResultQuestion result : resultQuestions) {
            try {
                lockService.watch(LockType.QUESTION, result.getQuestionId());
                Long currentTime = System.currentTimeMillis();
                updateMarkSubjectScore(markStudent, result, userId);
                markTaskService.updateHeaderResult(markStudent.getExamId(), markStudent.getPaperNumber(),
                        result.getQuestionId(), markStudent.getId(), userId, result.getMarkerScore(), result.getMarkerTrackList(), result.getMarkerTagList(), currentTime, MarkTaskStatus.MARKED);
                updateMarkedCount(markStudent.getExamId(), markStudent.getPaperNumber(), result.getQuestionId());
                markStudentService.updateCheckInfo(markStudent.getId(), userId);
                checkStudentQuestion(markStudent.getId(), markQuestionService.getById(result.getQuestionId()), result);
            } catch (ApiException e) {
                throw ExceptionResultEnum.ERROR.exception(e.getMessage());
            } finally {
                lockService.unwatch(LockType.QUESTION, result.getQuestionId());
            }
        }

        // 未分组的题目
//        checkStudentSubjective(markStudent.getId(), markStudent.getExamId(), markStudent.getPaperNumber());
    }

    private void updateMarkSubjectScore(MarkStudent markStudent, MarkResultQuestion resultQuestion, Long userId) {
        // 记录修改日志（按小题）
//        List<BasicOperationLog> basicOperationLogs = new ArrayList<>();
        MarkSubjectiveScore markSubjectiveScore = markSubjectiveScoreService.getByStudentIdAndQuestionId(markStudent.getId(), resultQuestion.getQuestionId());
        double score = resultQuestion.getMarkerScore();
        // 分数有变动，记录日志
        if (markSubjectiveScore.getScore() != score) {
//            MarkStudentVo markStudentVo = markStudentService.getDetailById(markStudent.getId());
//            String detail = String.format("%s（%s）课程%s试卷编号下，将%s（%s）的%s题从%s分修改为%s分", markStudentVo.getCourseName(), markStudentVo.getCourseCode(), markStudentVo.getPaperNumber(), markStudentVo.getStudentName(), markStudentVo.getStudentCode(), markSubjectiveScore.getMainNumber() + "-" + markSubjectiveScore.getSubNumber(), markSubjectiveScore.getScore(), score);
//            basicOperationLogs.add(new BasicOperationLog(Long.valueOf(ServletUtil.getRequestHeaderSchoolId().toString()), ServletUtil.getCurrentPrivilegeId(), OperationTypeEnum.SUBJECTIVE, OperationTypeEnum.SUBJECTIVE.getName(), ServletUtil.getRequest().getServletPath(), detail, String.valueOf(groupResult.getScore()), "成功", ServletUtil.getRequestUserId()));
            MarkHeaderHistory headerHistory = new MarkHeaderHistory();
            headerHistory.setId(SystemConstant.getDbUuid());
            headerHistory.setExamId(markStudent.getExamId());
            headerHistory.setPaperNumber(markStudent.getPaperNumber());
            headerHistory.setStudentId(markStudent.getId());
            headerHistory.setQuestionId(markSubjectiveScore.getQuestionId());
            headerHistory.setMainNumber(markSubjectiveScore.getMainNumber());
            headerHistory.setSubNumber(markSubjectiveScore.getSubNumber());
            headerHistory.setUserId(userId);
            headerHistory.setScore(score);
            headerHistory.setTrackList(JSON.toJSONString(resultQuestion.getMarkerTrackList()));
            headerHistory.setOriginalScore(markSubjectiveScore.getScore());
//            headerHistory.setOriginalTrackList();
            headerHistory.setCreateTime(System.currentTimeMillis());
            markHeaderHistoryService.save(headerHistory);
        }
        markSubjectiveScore.setScore(score);
        markSubjectiveScoreService.saveOrUpdateByMultiId(markSubjectiveScore);

        // 记录日志
//        if (CollectionUtils.isNotEmpty(basicOperationLogs)) {
//            basicOperationLogService.saveBatch(basicOperationLogs);
//        }
    }

    private String getKey(MarkStudent student) {
        return student.getExamId() + "_" + student.getPaperNumber();
    }

    @Override
    public IPage<Task> getHistory(Long userId, int pageNumber, int pageSize, Sort sort, String order, Long examId,
                                  String paperNumber, String secretNumber, Double markerScore) {
        if (!"marker_time".equals(order) && !"marker_score".equals(order)) {
            order = "marker_time";
        }
        Page<Long> page = new Page<>(pageNumber, pageSize);
        OrderItem orderItem = new OrderItem(order, sort.equals(Sort.asc));
        page.addOrder(orderItem);
        IPage<Long> list = markTaskService.listPageHistory(page, userId, examId, paperNumber, secretNumber, markerScore);
        List<Task> recordsDtos = new ArrayList<>();
        for (Long studentId : list.getRecords()) {
            List<MarkTask> markTaskList = markTaskService.listByStudentIdAndMarkerId(studentId, userId);
            Task dto = taskService.build(userId, markTaskList);
            recordsDtos.add(dto);
        }
        IPage<Task> result = new Page<>();
        result.setCurrent(list.getCurrent());
        result.setPages(list.getPages());
        result.setRecords(recordsDtos);
        result.setSize(list.getSize());
        result.setTotal(list.getTotal());
        return result;
    }

    @Override
    public Task getTask(Long userId, Long examId, String paperNumber, QuestionModel questionModel) {
        MarkPaper markPaper = markPaperService.getByExamIdAndPaperNumber(examId, paperNumber);
        if (markPaper == null) {
            throw ExceptionResultEnum.ERROR.exception("评卷试卷不存在");
        }
        int count = markUserQuestionService.countByExamIdAndPaperNumberAndAndUserId(examId, paperNumber, userId);
        if (count == 0) {
            throw ExceptionResultEnum.ERROR.exception("评卷任务被重置，请点击右上角返回按钮，重新在评卷入口菜单点击评卷");
        }
        Task task = null;
        List<Long> studentIds;
        int pageNumber = 1;
        while (task == null) {
//            if (markPaper.getClassMark()) {
//                studentIds = markTaskService.findUnMarkedFilterClass(new Page<>(pageNumber, 20), examId, paperNumber, userId, questionModel);
//            } else {
//                studentIds = markTaskService.findUnMarked(new Page<>(pageNumber, 20), examId, paperNumber, userId, questionModel);
//            }
            studentIds = markStudentService.findUnMarked(new Page<>(pageNumber, 20), examId, paperNumber, userId, markPaper.getClassMark());
            if (studentIds.isEmpty()) {
                break;
            }
            for (Long studentId : studentIds) {
                if (this.applyTask(examId, paperNumber, studentId, 1, userId)) {
                    Long questionId = null;
                    if (questionModel.equals(QuestionModel.SINGLE)) {
                        questionId = markTaskService.minQuestionIdByExamIdAndPaperNumber(examId, paperNumber, MarkTaskStatus.MARKED, MarkTaskStatus.ARBITRATED);
                    }
                    List<MarkTask> markTaskList = markTaskService.listByStudentIdAndUserId(studentId, userId, questionId);
                    task = taskService.build(userId, markTaskList);
                }
            }

            if (task == null) {
                pageNumber++;
            }
        }
        return task;
    }

    @Override
    public boolean applyTask(MarkTask t, Long userId) {
        // 查询待领取任务时，已经做了多评同一studentId互斥处理
        String key = String.valueOf(t.getStudentId());
        TaskLock taskLock = TaskLockUtil.getFormalTask(key);
        boolean lock = taskLock.add(t.getStudentId(), t.getTaskNumber(), userId);
        // 上锁失败直接返回
        if (!lock) {
            return false;
        }
        // 重复校验任务状态
        if (markTaskService.countByIdAndStatus(t.getId(), t.getStatus()) == 1) {
            return true;
        } else {
            taskLock.remove(t.getStudentId(), t.getTaskNumber(), userId);
            return false;
        }
    }

    @Override
    public boolean applyTask(Long examId, String paperNumber, Long studentId, Integer taskNumber, Long userId) {
        // 查询待领取任务时，已经做了多评同一studentId互斥处理
        String key = examId + "_" + paperNumber;
        TaskLock taskLock = TaskLockUtil.getFormalTask(key);
        boolean lock = taskLock.add(studentId, taskNumber, userId);
        // 上锁失败直接返回
        if (!lock) {
            return false;
        }
        // 重复校验任务状态
        if (markTaskService.countByStudentIdAndStatus(studentId, MarkTaskStatus.WAITING, MarkTaskStatus.REJECTED) > 0) {
            return true;
        } else {
            taskLock.remove(studentId, taskNumber, userId);
            return false;
        }
    }

    @Override
    public boolean hasApplied(MarkTask t, Long userId) {
        String key = t.getExamId() + "_" + t.getPaperNumber();
        TaskLock taskLock = TaskLockUtil.getFormalTask(key);
        return taskLock.exist(t.getStudentId(), t.getTaskNumber(), userId);
    }

    @Override
    public void deleteInitMarkData(Long examId, String paperNumber) {
        markPaperService.deleteByExamIdAndPaperNumber(examId, paperNumber);
        markStudentService.deleteByExamIdAndPaperNumber(examId, paperNumber);
        scanAnswerCardService.deleteByExamIdAndPaperNumber(examId, paperNumber);
        markQuestionService.deleteByExamIdAndPaperNumber(examId, paperNumber);
    }

    @Override
    public void calcObjectiveScore(MarkPaper markPaper) {
        int pageNumber = 1;
        int pageSize = 1000;
        IPage<MarkStudent> iPage = markStudentService.pageByExamAndPaperNumber(markPaper.getExamId(),
                markPaper.getPaperNumber(), pageNumber, pageSize);
        while (CollectionUtils.isNotEmpty(iPage.getRecords())) {
            for (MarkStudent student : iPage.getRecords()) {
                calculate(student);
            }
            pageNumber++;
            iPage = markStudentService.pageByExamAndPaperNumber(markPaper.getExamId(), markPaper.getPaperNumber(),
                    pageNumber, pageSize);
        }
    }

    private void calculate(MarkStudent student) {
        // 未上传、缺考的考生不统分
        if (!student.getUpload() || student.getAbsent() || student.getOmrAbsent() || (student.getManualAbsent() != null && student.getManualAbsent())) {
            return;
        }
        try {
            // 更新客观题得分
            markStudentService.calculateObjectiveScore(student);
            // 增加主观题总分统计
            // markService.scoreCalculate(student,
            // findMarkGroup(student.getSubjectCode()));
        } catch (Exception e) {
            log.error("calculate error for studentId=" + student.getId(), e);
        }
    }

    @Transactional
    @Override
    public SubmitResult saveTask(Long examId, String paperNumber, Long userId, MarkResult result) {
        for (MarkResultQuestion question : result.getQuestionList()) {
            MarkUserQuestion markUserQuestion = markUserQuestionService.getByExamIdAndPaperNumberAndQuestionIdAndUserId(examId, paperNumber, question.getQuestionId(), userId);
            if (markUserQuestion == null) {
                throw ExceptionResultEnum.ERROR.exception("评卷员未绑定题目" + question.getMainNumber() + "-" + question.getSubNumber());
            }
        }

        SubmitResult sr;
        try {
            lockService.watch(LockType.EXAM_SUBJECT, examId, paperNumber);
            lockService.watch(LockType.MARK_USER_QUESTION, userId);
            lockService.waitlock(LockType.STUDENT, result.getStudentId());
            sr = submitResult(examId, paperNumber, userId, result);
            if (sr.isSuccess()) {
                releaseTask(sr, userId);
            }
        } catch (Exception e) {
            log.error("save task error", e);
            throw ExceptionResultEnum.ERROR.exception(e.getMessage());
        } finally {
            lockService.unlock(LockType.STUDENT, result.getStudentId());
            lockService.unwatch(LockType.MARK_USER_QUESTION, userId);
            lockService.unwatch(LockType.EXAM_SUBJECT, examId, paperNumber);
        }

        if (sr == null || !sr.isSuccess()) {
            throw ExceptionResultEnum.ERROR.exception("评卷任务提交失败，请刷新页面");
        }
        return sr;
    }

    private SubmitResult submitResult(Long examId, String paperNumber, Long userId, MarkResult result) {
        int spentAvg = result.getSpent() / result.getQuestionList().size();
        int i = 0;
        for (MarkResultQuestion markResultQuestion : result.getQuestionList()) {
            try {
                lockService.watch(LockType.QUESTION, markResultQuestion.getQuestionId());
                markResultQuestion.setSpent(spentAvg);
                MarkUserQuestion markUserQuestion = markUserQuestionService.getByExamIdAndPaperNumberAndQuestionIdAndUserId(examId, paperNumber, markResultQuestion.getQuestionId(), userId);
                MarkQuestion markQuestion = markQuestionService.getById(markResultQuestion.getQuestionId());
                MarkTask task = markTaskService.getById(markResultQuestion.getTaskId());
                if (task != null && task.getExamId().equals(markUserQuestion.getExamId())
                        && task.getPaperNumber().equals(markUserQuestion.getPaperNumber())
                        && task.getQuestionId().equals(markUserQuestion.getQuestionId())) {
                    // 问题卷
                    Long now = System.currentTimeMillis();
                    if (markResultQuestion.isProblem()) {
                        // 状态更新
                        if (markTaskService.updateProblemResult(task.getId(), userId, now, markResultQuestion.getSpent())) {
                            saveProblemHistory(markResultQuestion, task, userId);
                            updateMarkedCount(markUserQuestion.getExamId(), markUserQuestion.getPaperNumber(), markUserQuestion.getQuestionId());
                            // 未评完
                            resetStudentGroup(task.getStudentId());
                            i++;
                        }
                    }
                    if (markResultQuestion.getMarkerScore() <= markQuestion.getTotalScore()) {//阅卷分是否小于等于该组总分
                        if (submitTask(task, userId, markQuestion, markResultQuestion)) {
                            updateMarkedCount(markUserQuestion.getExamId(), markUserQuestion.getPaperNumber(), markUserQuestion.getQuestionId());
                            i++;
                        }
                    }
                }
            } catch (Exception e) {
                throw ExceptionResultEnum.ERROR.exception(e.getMessage());
            } finally {
                lockService.unwatch(LockType.QUESTION, markResultQuestion.getQuestionId());
            }
        }
        if (CollectionUtils.size(result.getQuestionList()) == i) {
            return SubmitResult.success(result.getStudentId());
        } else {
            return SubmitResult.faile();
        }
    }

    private void saveProblemHistory(MarkResultQuestion result, MarkTask task, Long userId) {
        MarkProblemHistory history = markProblemHistoryService.findByTaskIdAndStatus(task.getId(),
                MarkProblemStatus.WAITING);
        if (history == null) {
            history = new MarkProblemHistory();
        }
        history.setId(SystemConstant.getDbUuid());
        history.setExamId(task.getExamId());
        history.setSecretNumber(task.getSecretNumber());
        history.setStudentCode(task.getStudentCode());
        history.setStudentId(task.getStudentId());
        history.setTaskId(task.getId());
        history.setPaperNumber(task.getPaperNumber());
        history.setQuestionId(task.getQuestionId());
        history.setType(result.getProblemType());
        if (MarkProblemType.OTHER.equals(result.getProblemType())) {
            history.setRemark(result.getProblemRemark());
        }
        history.setCreateTime(System.currentTimeMillis());
        history.setStatus(MarkProblemStatus.WAITING);
        history.setUserId(userId);
        markProblemHistoryService.save(history);
    }

    private boolean submitTask(MarkTask task, Long userId, MarkQuestion markQuestion, MarkResultQuestion result) {
        // 非本人领取的待评任务
        if ((task.getStatus() == MarkTaskStatus.WAITING || task.getStatus() == MarkTaskStatus.REJECTED)
                && !hasApplied(task, userId)) {
            return false;
        }
        // 非本人的回评任务
        if ((task.getStatus() == MarkTaskStatus.MARKED) && !task.getUserId().equals(userId)) {
            return false;
        }
        // 是否多评情况下已处理过该考生评卷任务
        if (markTaskService.countByStudentIdAndMarkerIdAndIdNotEqual(task.getStudentId(), task.getExamId(),
                task.getPaperNumber(), task.getQuestionId(), userId, task.getId()) > 0) {
            return false;
        }
        // 尝试提交评卷结果
        Long now = System.currentTimeMillis();
        //更新阅卷任务状态为已阅卷[阅卷分数，给分明细，阅卷时间，评卷时长]
        if (!markTaskService.updateMarkerResult(task.getId(), MarkTaskStatus.MARKED, userId,
                result, now, null, null, null, MarkTaskStatus.WAITING,
                MarkTaskStatus.MARKED, MarkTaskStatus.REJECTED)) {
            // 条件不符更新失败，直接返回
            return false;
        }
        // 判断多评模式下是否需要仲裁
        MarkArbitrateHistory history = null;
        if (markQuestion.getArbitrateThreshold() != null && markQuestion.getArbitrateThreshold() > 0) {
            // 多评模式
            List<MarkTask> list = markTaskService.findByStudentIdAndQuestionIdAndStatus(task.getStudentId(), task.getQuestionId(), MarkTaskStatus.MARKED);
            for (MarkTask other : list) {
                // 本评卷任务或组长已打分，则跳过该任务
                if (other.getId().equals(task.getId()) || other.getHeaderScore() != null) {
                    continue;
                }
                // 分差超过阀值
                if (Math.abs(other.getMarkerScore() - result.getMarkerScore()) > markQuestion.getArbitrateThreshold()) {
                    history = buildArbitrateHistory(task, now);
                    break;
                }
            }
        }
        if (history != null) {
            // 保存仲裁记录
            markArbitrateHistoryService.save(history);
            // 触发仲裁后续状态更新
            markTaskService.updateStatusByStudentIdAndQuestionId(task.getStudentId(), task.getQuestionId(), MarkTaskStatus.WAIT_ARBITRATE);
            // 未评完
            resetStudentGroup(task.getStudentId());
        } else {
            // 判断当前分组是否已完成评卷
            checkStudentQuestion(task.getStudentId(), markQuestion, result);
        }
        return true;
    }

    private MarkArbitrateHistory buildArbitrateHistory(MarkTask task, Long now) {
        MarkArbitrateHistory history = new MarkArbitrateHistory();
        history.setId(SystemConstant.getDbUuid());
        history.setExamId(task.getExamId());
        history.setPaperNumber(task.getPaperNumber());
        history.setQuestionId(task.getQuestionId());
        history.setStudentId(task.getStudentId());
        history.setStudentCode(task.getStudentCode());
        history.setSecretNumber(task.getSecretNumber());
        history.setStatus(MarkArbitrateStatus.WAITING);
        history.setCreateTime(now);
        return history;
    }

    @Override
    public void updateMarkGroupStatus(Long examId, String paperNumber) {
        // 更新MarkPaper中groupStatus
        long groupNumberIsNull = markQuestionService.countByExamIdAndPaperNumberAndObjectiveAndGroupNumberIsNull(examId,
                paperNumber, false);
        markPaperService.updateGroupStatusByExamIdAndPaperNumber(groupNumberIsNull == 0, examId, paperNumber);
    }

    @Override
    public void checkStudentSubjectiveScore(Long examId, String coursePaperId, long unGroupQuestionCount) {
        List<MarkStudent> markStudentList = markStudentService.listByExamIdAndCoursePaperId(examId, coursePaperId);
        for (MarkStudent markStudent : markStudentList) {
//            checkStudentSubjective(markStudent.getId(), unGroupQuestionCount, markQuestion.getPaperNumber());
        }
    }

    @Override
    @Transactional
    public boolean rejectMarkTask(MarkTask markTask, Long userId, String reason) {
        MarkPaper markPaper = markPaperService.getByExamIdAndPaperNumber(markTask.getExamId(), markTask.getPaperNumber());
        if (MarkPaperStatus.FINISH.equals(markPaper.getStatus())) {
            throw ExceptionResultEnum.MARK_PAPER_FINISH.exception();
        }
        MarkRejectHistory history = new MarkRejectHistory(markTask);
        MarkStudent markStudent = markStudentService.getById(markTask.getStudentId());
        history.setBasicStudentId(markStudent.getBasicStudentId());
        history.setRejectUserId(userId);
        history.setRejectReason(reason);
        Long now = System.currentTimeMillis();
        if (markTaskService.resetById(markTask.getId(), null, reason, userId, now, MarkTaskStatus.REJECTED)) {
            markUserQuestionService.updateRejectCountByExamIdAndPaperNumberAndQuestionIdAndUserId(markTask.getExamId(), markTask.getPaperNumber(), markTask.getQuestionId(), markTask.getUserId());
            markRejectHistoryService.save(history);
            markSubjectiveScoreService.updateRejected(markTask.getStudentId(), markTask.getQuestionId(), true);
            resetStudentGroup(markTask.getStudentId());
            updateMarkedCount(markTask.getExamId(), markTask.getPaperNumber(), markTask.getQuestionId());
            return true;
        } else {
            return false;
        }
    }

}
