package com.qmth.teachcloud.mark.service.impl;

import com.alibaba.fastjson.JSON;
import com.qmth.teachcloud.common.bean.marking.MarkConfigItem;
import com.qmth.teachcloud.common.contant.SystemConstant;
import com.qmth.teachcloud.common.entity.MarkQuestion;
import com.qmth.teachcloud.common.entity.SysUser;
import com.qmth.teachcloud.common.enums.mark.MarkProblemStatus;
import com.qmth.teachcloud.common.service.TeachcloudCommonService;
import com.qmth.teachcloud.common.util.ServletUtil;
import com.qmth.teachcloud.mark.dto.mark.MarkStudentVo;
import com.qmth.teachcloud.mark.dto.mark.ScoreItem;
import com.qmth.teachcloud.mark.dto.mark.SpecialTagDTO;
import com.qmth.teachcloud.mark.dto.mark.manage.MarkerTrackDTO;
import com.qmth.teachcloud.mark.dto.mark.manage.Task;
import com.qmth.teachcloud.mark.dto.mark.manage.TaskQuestion;
import com.qmth.teachcloud.mark.dto.mark.manage.TrackDTO;
import com.qmth.teachcloud.mark.entity.*;
import com.qmth.teachcloud.mark.enums.MarkTaskStatus;
import com.qmth.teachcloud.mark.service.*;
import com.qmth.teachcloud.mark.utils.BigDecimalUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

@Service
public class TaskServiceImpl implements TaskService {

    @Resource
    private MarkPaperService markPaperService;
    @Resource
    private MarkTaskService markTaskService;
    @Resource
    private MarkStudentService markStudentService;
    @Resource
    private MarkQuestionService markQuestionService;
    @Resource
    private MarkSubjectiveScoreService markSubjectiveScoreService;
    @Resource
    private TeachcloudCommonService teachcloudCommonService;
    @Resource
    private MarkProblemHistoryService markProblemHistoryService;

    @Override
    public Task build(MarkArbitrateHistory markArbitrateHistory, MarkQuestion markQuestion) {
        MarkStudent markStudent = markStudentService.getById(markArbitrateHistory.getStudentId());
        Task task = new Task();
        task.setStudentId(markArbitrateHistory.getStudentId());
        task.setSecretNumber(markArbitrateHistory.getSecretNumber());
        if (StringUtils.isNotBlank(markStudent.getObjectiveScoreList())) {
            task.setObjectiveScore(markStudent.getObjectiveScore());
        }
        if (StringUtils.isNotBlank(markStudent.getSheetPath())) {
            task.setSheetUrls(teachcloudCommonService.filePreview(markStudent.getSheetPathList(), SystemConstant.MARK_JPG_EXPIRE_MINUTE));
        }
        if (StringUtils.isNotBlank(markQuestion.getPicList())) {
            task.setSliceConfig(JSON.parseArray(markQuestion.getPicList(), MarkConfigItem.class));
        }
        task.setQuestionList(buildArbitrateStep(markQuestion, markArbitrateHistory));
        return task;
    }

    @Override
    public Task build(Long studentId) {
        MarkStudentVo markStudent = markStudentService.getMarkStudentVoByStudentId(studentId);
        Task task = new Task();
        if (markStudent != null) {
            task.setExamId(markStudent.getExamId());
            task.setStudentId(markStudent.getId());
            task.setSecretNumber(markStudent.getSecretNumber());
            task.setStudentCode(markStudent.getStudentCode());
            task.setStudentName(markStudent.getStudentName());
            task.setCourseCode(markStudent.getCourseCode());
            task.setCourseName(markStudent.getCourseName());
            task.setPaperNumber(markStudent.getPaperNumber());

            // 原图
            task.setSheetUrls(teachcloudCommonService.fileMarkPreview(markStudent.getSheetPathList()));

            List<MarkTask> markTaskList = markTaskService.listByStudentId(studentId);
            task.setSliceConfig(buildPic(markTaskList));
            task.setQuestionList(buildQuestionList(null, markTaskList));
            task.setMarkerScore(calcScore(markTaskList));
            MarkPaper markPaper = markPaperService.getByExamIdAndPaperNumber(markStudent.getExamId(), markStudent.getPaperNumber());
            task.setObjectiveScore(markPaper.getShowObjectScore() ? markStudent.getObjectiveScore() : null);
            return task;
        }
        return task;
    }

    private Double calcScore(List<MarkTask> markTaskList) {
        Double total = 0d;
        for (MarkTask markTask : markTaskList) {
            total = BigDecimalUtils.add(total, markTask.getMarkerScore());
        }
        return total;
    }

    /*private List<MarkStepDTO> buildMarkStep(MarkStudent markStudent) {
        Long examId = markStudent.getExamId();
        String paperNumber = markStudent.getPaperNumber();
        List<MarkStepDTO> list = new LinkedList<MarkStepDTO>();
        List<MarkQuestion> sList = markQuestionService.listByExamIdAndPaperNumberAndObjective(examId, paperNumber, false);
        Set<Integer> mainList = markSubjectiveScoreService.listMainNumberByStudentIdAndUncalculate(markStudent.getId(),
                true);
        List<MarkSubjectiveScore> scoreList = markSubjectiveScoreService.listByStudentIdAndQuestionId(markStudent.getId(), null);
        for (int i = 0; i < scoreList.size(); i++) {
            MarkQuestion question = sList.get(i);
            String questionNumber = question.getQuestionNumber();
            MarkStepDTO step = buildStep(question);
            if (!scoreList.isEmpty() && scoreList.size() == sList.size()) {
                step.setScore(scoreList.get(i).getScore());
                if (mainList.contains(step.getMainNumber())) {
                    step.setUncalculate(true);
                }
            }
            // 增加阅卷轨迹列表获取
            List<MarkTask> markTaskList = markTaskService.listByStudentIdAndQuestionId(markStudent.getId(), question.getId());
            // 仲裁轨迹
            List<TrackDTO> headerTracks = new ArrayList<>();
            for (MarkTask markTask : markTaskList) {
                headerTracks.addAll(markTask.listHeaderTrack());
            }
            Map<Long, String> userNameMap = new HashMap<>();
            // 不管单评还是多评显示所有评卷员给分轨迹
            for (MarkTask markTask : markTaskList) {
                List<TrackDTO> tracks = markTask.listMarkerTrack();
                for (TrackDTO track : tracks) {
                    if (userNameMap.get(track.getUserId()) == null) {
                        SysUser user = sysUserService.getByUserId(track.getUserId());
                        userNameMap.put(track.getUserId(), user.getRealName() + "(" + user.getLoginName() + ")");
                    }
                    step.addTrack(replaceTrackFromHeaderTrack(track, headerTracks, userNameMap));
                }

                // 普通模式无轨迹数据。单独处理评卷员每题得分
                if (CollectionUtils.isEmpty(tracks) && markTask.getMarkerScore() != null) {
                    // 评卷员
                    if (markTask.getMarkerScore() != null) {
                        MarkerScoreDTO markerScoreDTO = new MarkerScoreDTO();
                        markerScoreDTO.setUserId(markTask.getUserId());
                        SysUser sysUser = sysUserService.getByUserId(markTask.getUserId());
                        if (sysUser != null) {
                            markerScoreDTO.setLoginName(sysUser.getLoginName());
                            markerScoreDTO.setUserName(sysUser.getRealName());
                        }
                        markerScoreDTO.setScore(markTask.getMarkerScore());
                        markerScoreDTO.setHeader(false);
                        step.addMarkerList(markerScoreDTO);
                    }
                    // 科组长
                    if (markTask.getHeaderScore() != null) {
                        if (markTask.getMarkerScore() != markTask.getHeaderScore()) {
                            MarkerScoreDTO markerScoreDTO = new MarkerScoreDTO();
                            markerScoreDTO.setUserId(markTask.getHeaderId());
                            SysUser sysUser = sysUserService.getByUserId(markTask.getHeaderId());
                            if (sysUser != null) {
                                markerScoreDTO.setLoginName(sysUser.getLoginName());
                                markerScoreDTO.setUserName(sysUser.getRealName());
                            }
                            markerScoreDTO.setScore(markTask.getHeaderScore());
                            markerScoreDTO.setHeader(true);
                            step.addMarkerList(markerScoreDTO);
                        }
                    }
                }
            }

            for (MarkHeaderTrack headerTrack : headerTracks) {
                TrackDTO trackDTO = new TrackDTO(headerTrack);
                SysUser sysUser = sysUserService.getByUserId(headerTrack.getUserId());
                if (sysUser != null) {
                    trackDTO.setUserId(sysUser.getId());
                    trackDTO.setUserName(sysUser.getRealName() + "(" + sysUser.getLoginName() + ")");
                }
                step.addHeaderTrack(trackDTO);
            }
            list.add(step);
        }
        return list;
    }*/

    private MarkerTrackDTO replaceTrackFromHeaderTrack(MarkTrack markTrack, List<MarkHeaderTrack> headerTracks, Map<Long, String> userNameMap) {
        MarkerTrackDTO markerTrackDTO = new MarkerTrackDTO(markTrack, userNameMap.get(markTrack.getUserId()));
//        if (CollectionUtils.isNotEmpty(headerTracks)) {
//            Optional<MarkHeaderTrack> optional = headerTracks.stream().filter(m -> m.getStudentId().equals(markTrack.getStudentId()) && m.getQuestionNumber().equals(markTrack.getQuestionNumber())).findFirst();
//            if (optional.isPresent()) {
//                MarkHeaderTrack markHeaderTrack = optional.get();
//                SysUser user = sysUserService.getByUserId(markHeaderTrack.getUserId());
//                markerTrackDTO = new MarkerTrackDTO(markHeaderTrack, user.getRealName() + "(" + user.getLoginName() + ")");
//            }
//        }
        return markerTrackDTO;
    }

    private List<TaskQuestion> buildArbitrateStep(MarkQuestion markQuestion, MarkArbitrateHistory markArbitrateHistory) {
        List<TaskQuestion> list = new LinkedList<>();

        List<MarkTask> markTaskList = markTaskService.listByStudentIdAndQuestionId(markArbitrateHistory.getStudentId(), markArbitrateHistory.getQuestionId());
        List<ScoreItem> sItems = new ArrayList<ScoreItem>();
        TaskQuestion step = buildStep(markQuestion);
        if (markArbitrateHistory != null) {
            step.setScore(sItems.get(0).getScore());
            step.setMarkerTime(markArbitrateHistory.getUpdateTime());
            if (markArbitrateHistory.getTotalScore() != null) {
                step.setMarkerScore(markArbitrateHistory.getTotalScore());
            }
            // 增加阅卷轨迹列表获取
            List<TrackDTO> tracks = new ArrayList<>();
            for (MarkTask markTask : markTaskList) {
                tracks.addAll(markTask.listHeaderTrack());
            }
            step.setHeaderTrackList(tracks.toArray(new TrackDTO[tracks.size()]));
        }
        list.add(step);
        return list;
    }

    private TaskQuestion buildStep(MarkQuestion question) {
        TaskQuestion step = new TaskQuestion();
        step.setMainNumber(question.getMainNumber());
        step.setSubNumber(question.getSubNumber());
        step.setQuestionId(question.getId());
        step.setTitle(question.getMainTitle());
        step.setDefaultScore(0d);
        step.setMaxScore(question.getTotalScore());
        step.setMinScore(0d);
        step.setIntervalScore(question.getIntervalScore());
        step.setUncalculate(false);
        return step;
    }

    @Override
    public Task build(Long userId, List<MarkTask> markTaskList) {
        Long markStudentId = markTaskList.get(0).getStudentId();
        MarkStudent student = markStudentService.getById(markStudentId);
        Task task = new Task();
        task.setSecretNumber(student.getSecretNumber());
        task.setPaperNumber(student.getPaperNumber());
        task.setStudentId(markStudentId);
        task.setStudentCode(student.getStudentCode());
        task.setPaperType(student.getPaperType());
        task.setPaperType(student.getPaperType());
        task.setSheetUrls(teachcloudCommonService.fileMarkPreview(student.getSheetPathList()));
        task.setSliceConfig(buildPic(markTaskList));
        MarkPaper markPaper = markPaperService.getByExamIdAndPaperNumber(student.getExamId(), student.getPaperNumber());
        task.setObjectiveScore(markPaper.getShowObjectScore() ? student.getObjectiveScore() : null);
        // 题目数据
        task.setQuestionList(buildQuestionList(userId, markTaskList));
        return task;
    }

    private List<MarkConfigItem> buildPic(List<MarkTask> markTaskList) {
        List<MarkConfigItem> markConfigItems = new ArrayList<>();
        Integer mainNumber = null;
        for (MarkTask t : markTaskList) {
            MarkQuestion markQuestion = markQuestionService.getById(t.getQuestionId());
            if (markQuestion.getQuestionType() == 4) {
                // 填空题，只显示一次
                if (mainNumber == null || !markQuestion.getMainNumber().equals(mainNumber)) {
                    markConfigItems.addAll(markQuestion.getPictureConfigList());
                }
                mainNumber = markQuestion.getMainNumber();
            } else {
                markConfigItems.addAll(markQuestion.getPictureConfigList());
            }
        }
        return markConfigItems;
    }

    private List<TaskQuestion> buildQuestionList(Long userId, List<MarkTask> markTaskList) {
        List<TaskQuestion> taskQuestions = new ArrayList<>();
        for (MarkTask t : markTaskList) {
            TaskQuestion tq = new TaskQuestion();
            tq.setTaskId(t.getId());
            tq.setQuestionId(t.getQuestionId());
            tq.setMainNumber(t.getMainNumber());
            tq.setSubNumber(t.getSubNumber());
            MarkQuestion markQuestion = markQuestionService.getById(t.getQuestionId());
            tq.setIntervalScore(markQuestion.getIntervalScore());
            tq.setMaxScore(markQuestion.getTotalScore());
            tq.setMinScore(0);
            tq.setTitle(markQuestion.getMainTitle());
            tq.setStatus(t.getStatus());
            List<TrackDTO> markerTrack = t.listMarkerTrack();
            tq.setMarkerTrackList(markerTrack.toArray(new TrackDTO[markerTrack.size()]));
            List<SpecialTagDTO> markerTag = t.listMarkerTag();
            tq.setMarkerTagList(markerTag.toArray(new SpecialTagDTO[markerTag.size()]));
            List<TrackDTO> headerTrack = t.listHeaderTrack();
            tq.setHeaderTrackList(headerTrack.toArray(new TrackDTO[headerTrack.size()]));
            List<SpecialTagDTO> headerTag = t.listHeaderTag();
            tq.setHeaderTagList(headerTag.toArray(new SpecialTagDTO[headerTag.size()]));
            tq.setMarkerScore(t.getMarkerScore());
            tq.setSelfMark(userId == null || (t.getUserId() == null || t.getUserId().equals(userId)));
            tq.setRejected(t.getStatus().equals(MarkTaskStatus.REJECTED));
            tq.setPrevious(t.getStatus().equals(MarkTaskStatus.MARKED));
            tq.setProblem(t.getStatus().equals(MarkTaskStatus.PROBLEM));
            if (tq.isProblem()) {
                MarkProblemHistory markProblemHistory = markProblemHistoryService.findByTaskIdAndStatus(t.getId(), MarkProblemStatus.WAITING);
                if (markProblemHistory != null) {
                    tq.setProblemType(markProblemHistory.getType());
                    tq.setProblemRemark(markProblemHistory.getRemark());
                }
            }
            taskQuestions.add(tq);
        }
        return taskQuestions;
    }

    /*private List<MarkStepDTO> buildMarkStep(MarkTask task) {
        List<MarkStepDTO> list = new LinkedList<MarkStepDTO>();
        MarkQuestion markQuestion = markQuestionService.getById(task.getQuestionId());
        List<TrackDTO> tracks = task.listMarkerTrack();
        MarkStepDTO step = buildStep(markQuestion);
        if (task != null) {
            step.setScore(task.getMarkerScore());
            // 增加阅卷轨迹列表获取
            for (TrackDTO track : tracks) {
                step.addTrack(track);
            }
        }
        list.add(step);
        return list;
    }*/

    private SpecialTagDTO[] getHeaderTagList(MarkStudent student) {
        List<MarkTask> markTaskList = markTaskService.listByStudentId(student.getId());
        List<SpecialTagDTO> list = new ArrayList<>();
        for (MarkTask markTask : markTaskList) {
            list.addAll(markTask.listHeaderTag());
        }
//        SpecialTagDTO[] specialTags = new SpecialTagDTO[list.size()];
//        for (int i = 0; i < list.size(); i++) {
//            specialTags[i] = new MarkerSpecialTagDTO(list.get(i).getUserId(), list.get(i).getGroupNumber(), list.get(i));
//        }
        return list.toArray(new SpecialTagDTO[list.size()]);
    }

    public SpecialTagDTO[] getMarkSpecialTagList(MarkTask t) {
        SpecialTagDTO[] specialTags = null;
        List<SpecialTagDTO> list = t.listMarkerTag();
        if (!list.isEmpty()) {
            specialTags = list.toArray(new SpecialTagDTO[list.size()]);
        }
        return specialTags;
    }

    private SpecialTagDTO[] getMarkSpecialTagList(MarkStudent student) {
        List<SpecialTagDTO> list = new LinkedList<>();
        List<MarkTask> taskList = markTaskService.listByStudentId(student.getId());
        // 不管单评还是多评显示所有评卷员特殊标记
        for (MarkTask t : taskList) {
            List<SpecialTagDTO> tagList = t.listMarkerTag();
            list.addAll(tagList);
        }
        return list.toArray(new SpecialTagDTO[0]);
    }
}
